package ftcomputing.robo;

import java.lang.reflect.Field;

/**
 * Klasse zum Ansprechen der Fischertechnik-Interfaces ber Java. Basis ist die
 * umFish40.DLL v4.3.75.33 mit Wrapper.DLL javaFish40.DLL und der zugehrenden
 * Klasse JavaFish, die von der die Klasse FishFace abgeleitet ist. Der
 * Funktionsinhalt orientiert sich an der Klasse FishFace der C# Assembly
 * FishFace2005.DLL
 * 
 * @author Ulrich Mller
 * @author Michael Fennel
 * @see FishFaceException
 * @see JavaFish
 */
public class FishFace extends JavaFish {
	private final int jrError = 0xE0000001;
	private final int jrPollInterval = 12;
	private final int notOpen = 999;

	private int iHandle = notOpen;
	private boolean notHalt = true;
	private int distanceMode = Distance.Off;


	private String Name;
	private int Type;
	private int SerialNr;
	private String Firmware;
	
	/**
	 * Abfrage des Gertenamens (resultiert aus Gertetyp)
	 * 
	 * @see FishFace#Type
	 */
	public String getDeviceName(){
		if (iHandle == jrError)
			throw new FishFaceException("InterfaceProblem.OpenInterface");
		if(iHandle == notOpen)
			throw new FishFaceException("ConnectionProblem.NoConnection");
		return Name;
	}
	/**
	 * Abfrage des Devices-Typs
	 * @return Device-Typ (kongruent zu Gertename)
	 */
	public int getDeviceType(){
		if (iHandle == jrError)
			throw new FishFaceException("InterfaceProblem.OpenInterface");
		if(iHandle == notOpen)
			throw new FishFaceException("ConnectionProblem.NoConnection");
		return Type;
	}
	/**
	 * Abfrage der Seriennummer
	 * @return Seriennummer
	 */
	public int getDeviceSerial(){
		if (iHandle == jrError)
			throw new FishFaceException("InterfaceProblem.OpenInterface");
		if(iHandle == notOpen)
			throw new FishFaceException("ConnectionProblem.NoConnection");
		return SerialNr;
	}
	/**
	 * Abfrage der Firmware
	 * @return Firmware-Version
	 */
	public String getDeviceFirmware(){
		if (iHandle == jrError)
			throw new FishFaceException("InterfaceProblem.OpenInterface");
		if(iHandle == notOpen)
			throw new FishFaceException("ConnectionProblem.NoConnection");
		return Firmware;
	}

	/**
	 * Methode fr die Verbindung an einem USB-Port, Distance-Mode nicht einstellbar, steht aber Standardmig auf Ulta-Sonic
	 * 
	 * @see FishFace#openInterfaceUSB(int, int, int)
	 * @see FishFace#openInterfaceCOM(int, int, int)
	 * @param ifTyp
	 *            der Interface-Typ. Zu entnehmen aus der {@link IFTypen}
	 *            -Klasse.
	 * @param SerialNr
	 *            die Seriennummer des Interface. Bei
	 *            {@link IFTypen#ftROBO_first_USB} muss die Seriennummer
	 *            <code>0</code> sein
	 * @throws FishFaceException
	 *             wenn die Verbindung zum Interface fehl schlgt.
	 */
	public void openInterfaceUSB(int ifTyp, int SerialNr) throws FishFaceException {
		iHandle = jrOpenInterfaceUSB(ifTyp, SerialNr);
		if (iHandle == jrError)
			throw new FishFaceException("InterfaceProblem.OpenInterface");
		getActDevice();
		jrSetMotors(iHandle, 0);
		notHalt = false;
		distanceMode = Distance.UltraSonic;
	}

	/**
	 * Methode fr die Verbindung an einem USB-Port, mit einstellbarer
	 * Distance-Mode. Distance-Mode besser belassen, nur wer wei was er macht soll dies tun.
	 * @deprecated Distance-Mode sollte nicht manuell eingestellt werden!
	 * @see FishFace#openInterfaceUSB(int, int)
	 * @see FishFace#openInterfaceCOM(int, int, int)
	 * @param ifTyp
	 *            der Interface-Typ. Zu entnehmen aus der {@link IFTypen}
	 *            -Klasse.
	 * @param SerialNr
	 *            die Seriennummer des Interface. Bei
	 *            {@link IFTypen#ftROBO_first_USB} muss die Seriennummer
	 *            <code>0</code> sein
	 * @param DistanceMode
	 *            die Einstellung, wie die Distanzeingnge arbeiten sollen. Zu
	 *            entnehmen aus der {@link Distance}-Klasse.
	 * @throws FishFaceException
	 *             wenn die Verbindung zum Interface fehl schlgt.
	 * 
	 */
	public void openInterfaceUSB(int ifTyp, int SerialNr, int DistanceMode) {
		iHandle = jrOpenInterfaceUSBdis(ifTyp, SerialNr, DistanceMode);
		if (iHandle == jrError)
			throw new FishFaceException("InterfaceProblem.OpenInterface");
		getActDevice();
		jrSetMotors(iHandle, 0);
		notHalt = false;
		distanceMode = DistanceMode;
	}

	/**
	 * Methode fr die Verbindung an einem COM-Port. Distance-Mode nicht einstellbar, steht aber standardmig auf Ulta-Sonic.
	 * 
	 * @see FishFace#openInterfaceUSB(int, int)
	 * @see FishFace#openInterfaceUSB(int, int, int)
	 * @param ifTyp
	 *            der Interface-Typ. Zu entnehmen aus der {@link IFTypen}
	 *            -Klasse.
	 * @param ComNr
	 *            die Nummer des COM-Anschlusses am PC

	 * @param AnalogZyklen
	 *            gibt an nach wie vielen Zyklen die Werte der Analog-Eingnge
	 *            aktualisiert werden. Dies wurde vorgesehen, um die
	 *            Abfragefrequenz der brigen Eingnge zu erhhen, da die
	 *            Analogabfrage deutlich lnger dauert. Standard ist hier 5.
	 * @throws FishFaceException
	 *             wenn die Verbindung zum Interface fehl schlgt.
	 */
	public void openInterfaceCOM(int ifTyp, int ComNr, int AnalogZyklen) {
		iHandle = this.jrOpenInterfaceCOM(ifTyp, ComNr, AnalogZyklen);
		if (iHandle == jrError)
			throw new FishFaceException("InterfaceProblem.OpenInterface");
		getActDevice();
		jrSetMotors(iHandle, 0);
		notHalt = false;
		distanceMode = Distance.UltraSonic;
	}

	/**
	 * 
	 * @return die aktuelle Version der FishFace-Klasse
	 */
	public static String Version() {
		return "4.3.75.0";
	}

	/**
	 * Diese Methode dient dazu Informationen ber das Interface auszulesen.
	 * Wird im Konstruktor aufgerufen. Die Konstanten {@link FishFace#Name},
	 * {@link FishFace#Type}, {@link FishFace#SerialNr},
	 * {@link FishFace#Firmware} speichern die Daten ab. Sie knnen spter im
	 * Programm abgefragt werden.
	 * 
	 * @throws FishFaceException
	 *             wenn keine Verbindung zum Interface besteht oder es ein
	 *             Interface-Problem gibt.
	 */
	private void getActDevice() {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.getActDevice"));

		Type = jrGetActDeviceType(iHandle);
		SerialNr = jrGetActDeviceSerialNr(iHandle);
		IFTypen names = new IFTypen();
		Class<? extends IFTypen> c = names.getClass();
		Field[] felder = c.getFields();
		for (int i = 0; i < felder.length; i++) {
			try {
				if (felder[i].getInt(felder[i]) == Type) {
					Name = felder[i].getName();
				}
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		int f = jrGetActDeviceFirmwareNr(iHandle);
		Firmware = String.format("%d.%d.%d.%d", (f & 0xFF000000) >> 24,
				(f & 0xFF0000) >> 16, (f & 0xFF00) >> 8, (f & 0xFF));

	}

	/**
	 * Methode zum Abfragen eines Analogeingangs fr Wiederstnde
	 * <code>(AX, AY, AXS1 ... 3)</code>
	 * 
	 * @param AnalogNr
	 *            die Nummer des Analogeingangs. Siehe auch {@link Inp}.
	 * @return ein Wert zwischen 0 und 1023, desto grer der Wert, desto hher
	 *         ist der anliegende Wiederstand (der Messbereich reicht von
	 *         0-5,5k&#937;).
	 * 
	 */
	public int getAnalog(int AnalogNr) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.getAnalog"));
		int res = jrGetAnalog(iHandle, AnalogNr);
		if (res == jrError)
			throw (new FishFaceException("InterfaceProblem.getAnalog"));
		return res;
	}

	/**
	 * Abfrage des Counters eines Digitaleingangs
	 * 
	 * @param InputNr
	 *            die Nummer des Digitaleingangs, der abgefragt werden soll
	 *            (siehe {@link Inp}).
	 * @return die Zahl der Impulse, die zu dem gewhlten Eingang kamen. <br>
	 *         <b>Achtung:</b> Es wird jede Zustandsnderung am Digitaleingang
	 *         gezhlt. Dass heit, drckt man einen Taster einmal, so hat der
	 *         Counter 2 Impulse gezhlt.
	 *@throws FishFaceException
	 *             wenn keine Verbindung zum Interface besteht oder es ein
	 *             Interface-Problem gibt.
	 */
	public int getCounter(int InputNr) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.getCounter"));
		int res = jrGetCounter(iHandle, InputNr);
		if (res == jrError)
			throw (new FishFaceException("InterfaceProblem.getCounter"));
		return res;
	}

	/**
	 * Setzten eines bestimmten Counters auf einen bestimmten Wert.
	 * 
	 * @param InputNr
	 *            der Digitaleingang (siehe {@link Inp})
	 * @param ICount
	 *            die Zahl, auf die der Counter gestellt werden soll
	 */
	public void setCounter(int InputNr, int ICount) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.setCounter"));
		if (jrSetCounter(iHandle, InputNr, ICount) == jrError)
			throw (new FishFaceException("InterfaceProblem.setCounter"));
	}

	/**
	 * Auslesen der vom UltaSonic-Sensor gemessenen Entfernung in cm. Erstellen
	 * des FishFace-Objekts mit {@link Distance#UltraSonic} muss vorher gegangen
	 * sein.
	 * 
	 * @param SensorNr
	 *            die Nummer des Distanzsensoreingangs (siehe
	 *            {@link Distance#UltraSonic})
	 * @return gibt die Entfernung in cm zurck
	 */
	public int getDistance(int SensorNr) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.getDistance"));
		if (distanceMode != Distance.UltraSonic)
			throw (new FishFaceException("Kein DistanceMode.getDistance"));
		int res = jrGetDistanceValue(iHandle, SensorNr);
		if (res == jrError)
			throw (new FishFaceException("InterfaceProblem.getDistance"));
		return res;
	}

	/**
	 * Festellen des DistanceModes
	 * 
	 * @return die eingestellte DistanceMode als Zahl (entspricht
	 *         {@link Distance})
	 */
	public int getDistanceMode() {
		return distanceMode;
	}

	/**
	 * Abfrage des Zustands von einem I-Eingang
	 * 
	 * @param InputNr
	 *            Nummer des Digitaleingangs
	 * @return Zustand des Eingangs.
	 *         <ul>
	 *         <li>True: 1</li>
	 *         <li>False: 0</li>
	 */
	public boolean getInput(int InputNr) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.getInput"));
		int res = jrGetInput(iHandle, InputNr);
		if (res == jrError)
			throw (new FishFaceException("InterfaceProblem.getInput"));
		return (res == 1) ? true : false;
	}

	/**
	 * Abfrage des Zustands aller I-Eingnge
	 * 
	 * @return Die Summer aller I-Eingnge, diese errechnet sich wie folgt:<br>
	 *         <ul>
	 *         <li>
	 *         I1 hat den Wert 1, I2 hat den Wert 2, I3 hat den Wert 4, I4 hat
	 *         den Wert 8, usw...</li>
	 *         <li>
	 *         die Werte der I-Eingnge, die "geschlossen" sind, werden addiert</li>
	 *         </ul>
	 */
	public int getInputs() {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.getInputs"));
		int res = jrGetInputs(iHandle);
		if (res == jrError)
			throw (new FishFaceException("InterfaceProblem.getInputs"));
		return res;
	}

	/**
	 * Abfrage einer bestimmten Taste am IR-Handsender
	 * 
	 * @param Code
	 *            Auswahl des IR-Codes (siehe {@link IRCode})
	 * @param KeyNr
	 *            Auswahl der IR-Taste am Sender (siehe {@link IRKeys})
	 * @return true, falls die gewhlte Taste gedrckt ist,<br>
	 *         false, falls die gewhlte Taste nicht gedrckt ist
	 */
	public boolean getIRKey(int Code, int KeyNr) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.getIRKey"));
		int res = jrGetIRKey(iHandle, Code, KeyNr);
		if (res == jrError)
			throw (new FishFaceException("InterfaceProblem.getIRKey"));
		if (res == 1)
			return true;
		else
			return false;
	}

	/**
	 * Schalten einer Lampe an einem O-Ausgang (nur Robo-Interface)
	 * 
	 * @param LampNr
	 *            die Nummer des Ausgangs (siehe {@link Out})
	 * @param OnOff
	 *            der Zustand, der gesetzt werden soll (siehe {@link Dir})
	 */
	public void setLamp(int LampNr, int OnOff) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.setLamp"));
		if (jrSetLamp(iHandle, LampNr, OnOff) == jrError)
			throw (new FishFaceException("InterfaceProblem.setLamp"));
	}

	/**
	 * Schalten einer Lampe an einem O-Ausgang (nur Robo-Interface) mit Angabe
	 * der Helligkeit
	 * 
	 * @param LampNr
	 *            die Nummer des Ausgangs (siehe {@link Out})
	 * @param OnOff
	 *            der Zustand, der gesetzt werden soll (siehe {@link Dir})
	 * @param Brightness
	 *            die Helligkeit der Lampe (siehe {@link Speed})
	 */
	public void setLamp(int LampNr, int OnOff, int Brightness) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.setLamp"));
		if (jrSetLampEx(iHandle, LampNr, OnOff, Brightness) == jrError)
			throw (new FishFaceException("InterfaceProblem.setLamp"));
	}

	/**
	 * Setzen eines M-Ausganges
	 * 
	 * @param MotorNr
	 *            die Nummer des Motors (siehe {@link Mot})
	 * @param Direction
	 *            die Drehrichtung, die gesetzt werden soll (siehe {@link Dir})
	 */
	public void setMotor(int MotorNr, int Direction) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.setMotor"));
		if (jrSetMotor(iHandle, MotorNr, Direction) == jrError)
			throw (new FishFaceException("InterfaceProblem.setMotor"));
	}

	/**
	 * Setzen eines M-Ausganges mit Angabe der Geschwindigkeit
	 * 
	 * @param MotorNr
	 *            die Nummer des Motors (siehe {@link Mot})
	 * @param Direction
	 *            die Drehrichtung, die gesetzt werden soll (siehe {@link Dir})
	 * @param Speed
	 *            die Geschwindigkeit des Motors (siehe {@link Speed})
	 */
	public void setMotor(int MotorNr, int Direction, int Speed) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.setMotor"));
		if (jrSetMotorEx(iHandle, MotorNr, Direction, Speed) == jrError)
			throw (new FishFaceException("InterfaceProblem.setMotor"));
	}

	/**
	 * Setzen eines M-Ausgangs mit Angabe der Geschwindigkeit und bei
	 * angeschlossenen Impulszhler zum Setzen der zu drehenden Impulse. Der
	 * Impulszhler des jeweiligen Motor ist am I-Eingang
	 * <code>MotorNr * 2</code>. Jeder Wechsel des Zustandes stellt einen Impuls
	 * dar. Auerdem gibt es eien Stopp-Taster am I-Eingang
	 * <code>MotorNr * 2 -1</code>.
	 * 
	 * @param MotorNr
	 *            die Nummer des Motors (siehe {@link Mot})
	 * @param Direction
	 *            die Drehrichtung, die gesetzt werden soll (siehe {@link Dir})
	 * @param Speed
	 *            die Geschwindigkeit des Motors (siehe {@link Speed})
	 * @param ICount
	 *            Zahl der zu drehenden Impulse
	 */
	public void setMotor(int MotorNr, int Direction, int Speed, int ICount) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.setMotor"));
		if (jrRobMotor(iHandle, MotorNr, Direction, Speed, ICount) == jrError)
			throw (new FishFaceException("InterfaceProblem.setMotor"));
	}

	/**
	 * Setzen aller M-Ausgnge
	 * 
	 * @param MotorStatus
	 *            die Drehrichtung, die gesetzt werden soll (siehe {@link Dir})
	 */
	public void setMotors(int MotorStatus) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.setMotors"));
		if (jrSetMotors(iHandle, MotorStatus) == jrError)
			throw (new FishFaceException("InterfaceProblem.setMotors"));
	}

	/**
	 * Setzen aller M-Ausgnge mit Angabe der Geschwindigkeit
	 * 
	 * @param MotorStatus
	 *            die Drehrichtung, die gesetzt werden soll (siehe {@link Dir})
	 * @param Speed
	 *            die Geschwindigkeit des Motors (siehe {@link Speed})
	 */
	public void setMotors(int MotorStatus, int Speed) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.setMotors"));
		if (jrSetMotorsEx(iHandle, MotorStatus, Speed, Speed) == jrError)
			throw (new FishFaceException("InterfaceProblem.setMotors"));
	}

	/**
	 * Abfragen des notHalt-Status
	 * 
	 * @return false, wenn kein Nothalt vorliegt<br>
	 *         true, wenn Nothalt vorliegt
	 */
	public boolean getNotHalt() {
		return notHalt;
	}

	/**
	 * Setzen des notHalt-Status
	 * 
	 * @param OnOff
	 *            Status, der gesetzt werden soll
	 */
	public void setNotHalt(boolean OnOff) {
		notHalt = OnOff;
	}

	/**
	 * Auslesen des Status aller M-Ausgnge
	 * 
	 * @return Die Summer aller M/O-Ausgnge, diese errechnet sich wie folgt:<br>
	 *         <ul>
	 *         <li>
	 *         O1 hat den Wert 1, O2 hat den Wert 2, O3 hat den Wert 4, O4 hat
	 *         den Wert 8, usw..., wenn der Zustand "an" ist</li>
	 *         <li>M1 mit Drehrichtung links hat den Wert 1, mir Drehrichtung
	 *         rechts 2, M2 mit Drehrichtung links hat den Wert 4, mit
	 *         Drehrichtung rechts 8</li>
	 *         <li>
	 * 
	 *         alle Werte werden addiert</li>
	 *         </ul>
	 */
	public int getOutputs() {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.getOutputs"));
		int res = jrGetMotors(iHandle);
		if (res == jrError)
			throw (new FishFaceException("InterfaceProblem.getOutputs"));
		return res;
	}

	/**
	 * Abfragen der analogen Spannungseingnge bzw. der Distanzeingngen im
	 * Voltage-Mode ({@link Distance#Voltage})
	 * 
	 * @param VoltNr
	 *            Nummer des Spannungseingangs ({@link Inp})
	 * @return einen Wert von 0-1023 (entstpricht 0 - 10 V)
	 */
	public int getVoltage(int VoltNr) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.GetVoltage"));
		int res = jrGetVoltage(iHandle, VoltNr);
		if (res == jrError)
			throw (new FishFaceException("InterfaceProblem.GetVoltage"));
		return res;
	}

	/**
	 * Zurcksetzen eines Counters auf 0
	 * 
	 * @param InputNr
	 *            der Counter, der auf 0 gesetzt werden soll (siehe {@link Inp})
	 */
	public void clearCounter(int InputNr) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.ClearCounter"));
		if (jrSetCounter(iHandle, InputNr, 0) == jrError)
			throw (new FishFaceException("InterfaceProblem.ClearCounter"));
	}

	/**
	 * Zrucksetzen aller Counter auf 0
	 */
	public void clearCounters() {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.ClearCounters"));
		if (jrClearCounters(iHandle) == jrError)
			throw (new FishFaceException("InterfaceProblem.ClearCounters"));
	}

	/**
	 * Lschen aller M-Ausgnge (damit werden auch alle O-Ausgngen gelscht)
	 */
	public void clearMotors() {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.ClearMotors"));
		if (jrSetMotors(iHandle, 0) == jrError)
			throw (new FishFaceException("InterfaceProblem.ClearMotors"));
	}

	/**
	 * Beenden der Verbindungn zum Interface
	 */
	public void closeInterface() {
		if (iHandle != notOpen)
			if (jrCloseInterface(iHandle) == jrError)
				throw (new FishFaceException("InterfaceProblem.CloseInterface"));
		iHandle = notOpen;
		distanceMode = Distance.Off;
	}

	/**
	 * Feststellen eines Endewunsches durch die Escape-Taste oder das
	 * notHalt-Feld.
	 * 
	 * @return true, wenn die Escape-Taste gedrckt wird oder notHalt == true
	 *         ist.<br>
	 *         false, wenn kein Endewunsch besteht
	 */
	public boolean finish() {
		return finish(0);
	}

	/**
	 * Feststellen eines Endewunsches durch die Escape-Taste, einen bestimmten
	 * Taster oder das notHalt-Feld.
	 * 
	 * @param InputNr
	 *            die Nummer des I-Eingangs ({@link Inp})
	 * @return true, wenn die Escape-Taste gedrckt wird, der gewhlte Taster
	 *         gedrckt oder notHalt == true ist<br>
	 *         false, wenn kein Endewunsch besteht
	 */
	public boolean finish(int InputNr) {
		Thread.yield();
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.Finish"));
		if (notHalt || escape() != 0)
			return true;
		if (jrGetInputs(iHandle) == jrError)
			throw (new FishFaceException("InterfaceProblem.Finish"));
		if (InputNr > 0 && getInput(InputNr))
			return true;
		return false;
	}

	/**
	 * Anhalten des Programms, kann durch notHalt == true oder Escape
	 * unterbrochen werden
	 * 
	 * @param mSek
	 *            Zeit in Millisekunden, die angibt wie lange das Programm
	 *            angehalten werden soll
	 */
	public void pause(int mSek) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.Pause"));
		int sZeit = getTickCount();
		do {
			Thread.yield();
			if (notHalt || escape() != 0)
				return;
			if (jrGetInputs(iHandle) == jrError)
				throw (new FishFaceException("InterfaceProblem.WaitForInput"));
			sleep(jrPollInterval);
		} while (getTickCount() - sZeit < mSek);
	}

	/**
	 * Warten auf einen angebene Zahl von Zustandsnderungen an einem I-Eingang,
	 * unterbrechbar
	 * 
	 * @param InputNr
	 *            die Nummer des I-Eingangs ({@link Inp})
	 * @param NrOfChanges
	 *            Zahl der Zustandsnderungen
	 * @return gibt an, wie die Methode beendet wurde: NotHalt, Escape-Taste,
	 *         Abbruchtaster, Zahl der Zustandsnderungen erreicht (zu entnehmen
	 *         aus der {@link Wait}-Klasse)
	 */
	public int waitForChange(int InputNr, int NrOfChanges) {
		return waitForChange(InputNr, NrOfChanges, 0);
	}

	/**
	 * Warten auf eine beliebige Zahl von Zustandsnderungen an einem I-Eingang,
	 * kann durch drcken eines Taster abgebrochen werden
	 * 
	 * @param InputNr
	 *            die Nummer des I-Eingangs ({@link Inp})
	 * @param NrOfChanges
	 *            Zahl der Zustandsnderungen
	 * @param TermInputNr
	 *            die Nummer des I-Eingangs ({@link Inp}), der fr den Abbruch
	 *            sorgt (bei 0 gibt es keine I-Eingang, der abbricht)
	 * @return gibt an, wie die Methode beendet wurde: NotHalt, Escape-Taste,
	 *         Abbruchtaster, Zahl der Zustandsnderungen erreicht (zu entnehmen
	 *         aus der {@link Wait}-Klasse)
	 */
	public int waitForChange(int InputNr, int NrOfChanges, int TermInputNr) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.WaitForChange"));
		jrSetCounter(iHandle, InputNr, 0);
		while (jrGetCounter(iHandle, InputNr) < NrOfChanges) {
			if (notHalt)
				return Wait.NotHalt;
			if (escape() != 0)
				return Wait.ESC;
			if (TermInputNr > 0)
				if (getInput(TermInputNr))
					return Wait.Abort;
			if (jrGetInputs(iHandle) == jrError)
				throw (new FishFaceException("InterfaceProblem.WaitForInput"));
			sleep(jrPollInterval);
			Thread.yield();
		}
		return Wait.End;
	}

	/**
	 * Warten auf einen Wechsel des Zustandes am angebgenen I-Eingang,
	 * unterbrechbar
	 * 
	 * @param InputNr
	 *            die NUmmer des I-Eingangs ({@link Inp})
	 * @return gibt an, wie die Methode beendet wurde: NotHalt, Escape-Taste,
	 *         Abbruchtaster, Zahl der Zustandsnderungen erreicht (zu entnehmen
	 *         aus der {@link Wait}-Klasse)
	 */
	public int waitForChanges(int InputNr) {
		return waitForChange(InputNr, 1, 0);
	}

	/**
	 * Warten auf einen Wechsel von 0 nach 1 am angegebenen I-Eingang
	 * 
	 * @param InputNr
	 *            die Nummer des I-Eingangs ({@link Inp})
	 * @return gibt an, wie die Methode beendet wurde: NotHalt, Escape-Taste,
	 *         Abbruchtaster, Zahl der Zustandsnderungen erreicht (zu entnehmen
	 *         aus der {@link Wait}-Klasse)
	 */
	public int waitForHigh(int InputNr) {

		int ret;
		ret = waitForInput(InputNr, false);
		if (ret != Wait.End) {
			return ret;
		}
		ret = waitForInput(InputNr, true);
		return ret;
	}

	/**
	 * Warten auf true (1) am angegebenen I-Eingang, abbrechbar
	 * 
	 * @param InputNr
	 *            die Nummer des I-Eingangs ({@link Inp})
	 * @return gibt an, wie die Methode beendet wurde: NotHalt, Escape-Taste,
	 *         Abbruchtaster, Zahl der Zustandsnderungen erreicht (zu entnehmen
	 *         aus der {@link Wait}-Klasse)
	 */
	public int waitForInput(int InputNr) {
		return waitForInput(InputNr, true);
	}

	/**
	 * Warten auf einen Wechsel von 1 nach 0 am angebenen I-Eingang, abbrechbar
	 * 
	 * @param InputNr
	 *            die Nummer des I-Eingangs ({@link Inp})
	 * @return gibt an, wie die Methode beendet wurde: NotHalt, Escape-Taste,
	 *         Abbruchtaster, Zahl der Zustandsnderungen erreicht (zu entnehmen
	 *         aus der {@link Wait}-Klasse)
	 */
	public int waitForLow(int InputNr) {
		int ret;
		ret = waitForInput(InputNr, true);
		if (ret != Wait.End) {
			return ret;
		}
		ret = waitForInput(InputNr, false);
		return ret;
	}

	/**
	 * Warten auf 0 oder 1 am angegebenen I-Eingang, abbrechbar
	 * 
	 * @param InputNr
	 *            die Nummer des I-Eingangs ({@link Inp})
	 * @param OnOff
	 *            <li>true: warten auf 1</li><li>false: warten auf 0</li>
	 * @return gibt an, wie die Methode beendet wurde: NotHalt, Escape-Taste,
	 *         Abbruchtaster, Zahl der Zustandsnderungen erreicht (zu entnehmen
	 *         aus der {@link Wait}-Klasse)
	 */
	public int waitForInput(int InputNr, boolean OnOff) {
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.WaitForInput"));
		do {
			if (notHalt)
				return Wait.NotHalt;
			if (escape() != 0)
				return Wait.ESC;
			if (jrGetInputs(iHandle) == jrError)
				throw (new FishFaceException("InterfaceProblem.WaitForInput"));
			sleep(jrPollInterval);
			Thread.yield();
		} while (!getInput(InputNr) == OnOff);
		return Wait.End;
	}

	/**
	 * Warten auf ein Motor-Ready-Ereignis der angegebenen M-Ausgnge
	 * {@link Mot}.M1... Die M-Ausgnge mssen als RobMotoren mit
	 * {@link FishFace#setMotor(int, int, int, int)} einzeln gestartet werden.
	 * Das Ready-Ereignis tritt ein, wenn alle zugehrenden Counter auf 0
	 * stehen. Die Methode kann in einer Schleife ohne erneuten Motorstart
	 * wiederholt werden. Der Return-Wert gibt an, wie die Methode beendet
	 * wurde. Zu entnehmen aus der {@link Wait}-Klasse.
	 * 
	 * @param mSek
	 *            gibt an, wie lange auf das Ready-Ereignis gewartet werden
	 *            soll. 0 Endlos, sonst die angegebene Anzahl von MilliSekunden.
	 * @param MotorNrs
	 *            die Motoren, auf die gewartet werden soll
	 * @return gibt an, wie die Methode beendet wurde. Zu entnehmen aus der
	 *         {@link Wait}-Klasse.
	 */
	public int waitForMotors(int mSek, int... MotorNrs) {
		int state;
		if (iHandle == notOpen)
			throw (new FishFaceException("KeinOpen.WaitForMotors"));
		int time = getTickCount() + mSek;
		do {
			state = 0;
			for (int motNr : MotorNrs) {
				state += jrGetModeStatus(iHandle, motNr);
			}
			if (state == 0)
				return Wait.End;
			if (notHalt || (escape() != 0))
				for (int motNr : MotorNrs)
					jrSetMotor(iHandle, motNr, 0);
			if (notHalt)
				return Wait.NotHalt;
			if (escape() != 0)
				return Wait.ESC;
			Thread.yield();
			sleep(jrPollInterval);
			if (jrGetInputs(iHandle) == jrError)
				throw (new FishFaceException("InterfaceProblem.WaitForMotors"));
		} while ((mSek == 0) || time > getTickCount());
		return Wait.Time;
	}
}
